@@ -24,8 +24,8 @@
 #include "report.hpp"                   // for REPORT
 #include "timer.hpp"                    // for detail::Timer
 
-#ifndef LIBSEMIGROUPS_INCLUDE_FROIDURE_PIN_IMPL_HPP_
-#define LIBSEMIGROUPS_INCLUDE_FROIDURE_PIN_IMPL_HPP_
+#ifndef LIBSEMIGROUPS_FROIDURE_PIN_IMPL_HPP_
+#define LIBSEMIGROUPS_FROIDURE_PIN_IMPL_HPP_
 
 #define TEMPLATE template <typename TElementType, typename TTraits>
 #define FROIDURE_PIN FroidurePin<TElementType, TTraits>
@@ -423,7 +423,7 @@ namespace libsemigroups {
   }
 
   ELEMENT_INDEX_TYPE FROIDURE_PIN::current_position(const_reference x) const {
-    if (Degree()(this->to_internal_const(x)) != _degree) {
+    if (Degree()(x) != _degree) {
       return UNDEFINED;
     }
 
@@ -524,8 +524,10 @@ namespace libsemigroups {
   }
 
   BOOL FROIDURE_PIN::is_idempotent(element_index_type pos) {
-    validate_element_index(pos);
     init_idempotents();
+    // only validate pos after init_idempotents, because we don't know if it's
+    // valid until then
+    validate_element_index(pos);
     return _is_idempotent[pos];
   }
 
@@ -615,7 +617,9 @@ namespace libsemigroups {
   }
 
   ELEMENT_INDEX_TYPE FROIDURE_PIN::right(element_index_type i, letter_type j) {
+    validate_letter_index(j);
     run();
+    validate_element_index(i);
     return _right.get(i, j);
   }
 
@@ -626,7 +630,9 @@ namespace libsemigroups {
   }
 
   ELEMENT_INDEX_TYPE FROIDURE_PIN::left(element_index_type i, letter_type j) {
+    validate_letter_index(j);
     run();
+    validate_element_index(i);
     return _left.get(i, j);
   }
 
@@ -872,7 +878,7 @@ namespace libsemigroups {
   }
 
   BOOL FROIDURE_PIN::finished_impl() const {
-    return _pos >= _nr;
+    return !running() && _pos >= _nr;
   }
 
   VOID FROIDURE_PIN::add_generator(element_type const& x) {
@@ -1317,7 +1323,7 @@ namespace libsemigroups {
     }
     _idempotents_found = true;
     run();
-    _is_idempotent.resize(_nr, false);
+    _is_idempotent.resize(_nr, 0);
 
     detail::Timer timer;
 
@@ -1446,7 +1452,7 @@ namespace libsemigroups {
 
     for (; pos < std::min(threshold, last); pos++) {
       element_index_type k = _enumerate_order[pos];
-      if (!_is_idempotent[k]) {
+      if (_is_idempotent[k] == 0) {
         // The following is product_by_reduction, don't have to consider
         // lengths because they are equal!!
         element_index_type i = k, j = k;
@@ -1459,7 +1465,7 @@ namespace libsemigroups {
         }
         if (i == k) {
           idempotents.emplace_back(_elements[k], k);
-          _is_idempotent[k] = true;
+          _is_idempotent[k] = 1;
         }
       }
     }
@@ -1475,14 +1481,14 @@ namespace libsemigroups {
 
     for (; pos < last; pos++) {
       element_index_type k = _enumerate_order[pos];
-      if (!_is_idempotent[k]) {
+      if (_is_idempotent[k] == 0) {
         Product()(this->to_external(tmp_product),
                   this->to_external(_elements[k]),
                   this->to_external(_elements[k]),
                   tid);
         if (InternalEqualTo()(tmp_product, _elements[k])) {
           idempotents.emplace_back(_elements[k], k);
-          _is_idempotent[k] = true;
+          _is_idempotent[k] = 1;
         }
       }
     }
@@ -1605,4 +1611,4 @@ namespace libsemigroups {
   };
 
 }  // namespace libsemigroups
-#endif  // LIBSEMIGROUPS_INCLUDE_FROIDURE_PIN_IMPL_HPP_
+#endif  // LIBSEMIGROUPS_FROIDURE_PIN_IMPL_HPP_
